' Minimalist Z80 Demo
CON
  _clkmode = xtal1 + pll16x
  _xinfreq = 5_000_000
  _baudRateSpeed = 115200 ' Teraterm does not work at this speed. Use hyperterminal or proptool terminal
  _receiverPin = 31
  _transmitterPin = 30

OBJ
  delay: "Timing"                                       ' for millisecond delays
  sio  : "pcFullDuplexSerial2FC"                        ' serial port
  str  : "ASCII0_STREngine"                             ' strings

VAR

PUB Main
   sio.AddPort(0, _receiverPin, _transmitterPin, -1, -1, 0, 0, _baudRateSpeed)
   sio.start
   delay.pause1ms(1000) ' time to press F12 to start the terminal
   PrintStringCR(string("start"))
   
   Z80Boot
   repeat  ' endless loop

PUB Z80Boot | i
   ' P0-7 is Z80 data bus 0-7
   ' P8 = Z80 /IORQ
   ' P9 = Z80 /WR
   ' P10 = Z80 /RD
   ' P11 = Z80 Clk
   ' P12 = Z80 A0
   ' P13 = Z80 /RESET
   Z80ResetLow                                          ' set reset low
   Z80Clock(5)                                          ' clock several times to register the reset
   Z80ResetHigh                                         ' set reset high
   Z80WaitForRdLow                                      ' clock until rd goes low, at this point address should be 00000000_00000000
   repeat i from 0 to 63                                ' make sure number is bigger than DAT program                           
      Z80WriteByteToRam(byte[@Z80BootProgram][i])       ' move a byte into ram
      Z80DataBusNOP                                     ' sets data bus to zero, Z80 reads this as a NOP and proceeds to next instruciton
      Z80Clock(2)                                       ' two clocks
      Z80WaitForRdLow                                   ' wait until reads next value
   Z80ResetLow                                          ' reset the Z80 ready to run the program
   Z80Clock(5)                                          ' wait till responds
   Z80ResetHigh                                         ' reset high
   'Z80Clock(3)                                          'run the little program, test pin status with different number of clock pulses

   ReceiveBlock($0000)                                ' read back the program
   SendBlock($0100,$55)                               ' send a block of data to the Z80
   SendBlock($0200,$AA)                               ' send a block of data to the Z80 
   ReceiveBlock($0100)
   ReceiveBlock($0200)                                ' read the block back

   ' test the speed, send 8 blocks of 128 bytes = 1024 bytes
   crlf
   PrintStringCR(string("Write 8192 bytes"))
   repeat 64
     SendBlock($0100,$44)
     PrintChar(".")
   PrintStringCR(string("Finished"))      
   repeat  ' do nothing

      
DAT
   Z80BootProgram
        byte $31,$FF,$FF,$DB,$01,$FE,$80,$CA,$17,$00,$FE,$81,$CA,$25,$00,$FE
        byte $82,$CA,$3A,$00,$C3,$03,$00,$CD,$33,$00,$06,$80,$DB,$00,$77,$23 
        byte $10,$FA,$C3,$03,$00,$CD,$33,$00,$06,$80,$7E,$D3,$00,$23,$10,$FA 
        byte $C3,$03,$00,$DB,$00,$6F,$DB,$00,$67,$C9,$C3,$F0,$FF 

{        
; minimalist bootstrap program - routines to transfer data to and from memory
; port 0 is data and port 1 is commands
; for something like CP/M that sits in low memory, need to get a bootstrap
; program running in high memory so can transfer CP/M as one memory image
; scan port 1 for a command as below:
; number        command
; 128           block move 128 bytes from propeller to Z80 ram 
; 129           block move 128 bytes from Z80 to propeller
; 130           jump to location 0FFF0H

                .Z80
                org 0H
start:          ld      sp,0FFFFH               ; set stack pointer to top of ram for calls
main:           in      a,(1)                   ; read command byte from port 1
                cp      128                     ; test the command byte
                jp      z,proptoz80             ; data from prop to z80
                cp      129
                jp      z,z80toprop             ; data from Z80 to prop
                cp      130             
                jp      z,highmem
                jp      main                    ; keep testing

proptoz80:      call    getaddress              ; get two bytes from the propeller and put in HL
                ld      b,128                   ; set up counter for 128 bytes to transfer
loop1:          in      a,(0)                   ; get byte from data port
                ld      (hl),a          ; store to memory
                inc     hl                      ; increment memory counter
                djnz    loop1                   ; do 128 times
                jp      main

z80toprop:      call    getaddress              ; get two bytes from the propeller and put in HL
                ld      b,128                   ; set up counter for 128 bytes to transfer
loop2:          ld      a,(hl)          ; get byte from memory
                out     (0),a                   ; send to the propeller
                inc     hl                      ; increment memory counter
                djnz    loop2                   ; do 128 times  
                jp      main                    ; return to main loop

getaddress:     in      a,(0)                   ; collect lsb
                ld      l,a                     ; move to register l
                in      a,(0)                   ; collect msb
                ld      h,a                     ; move to register h
                ret

highmem:        jp      0FFF0H          ; jump to a high memory location                                
                end
}
   

PUB Z80Clock(n)
    DIRA |= %00000000_00000000_00001000_00000000        ' enable the clock pin as an output\
    repeat n
      OUTA |= %00000000_00000000_00001000_00000000        ' clock high
      'delay.pause1ms(1)                                ' delay if needed
      OUTA &= %11111111_11111111_11110111_11111111        ' clock low
      'delay.pause1ms(1)                                ' delay

PUB Z80ResetLow
   DIRA &= %11111111_11111111_11000000_00000000         ' data bus and /iorq /wr /rd clk A0 and /reset all HiZ (pulled high with 10k)
   DIRA |= %00000000_00000000_00100000_00000000         ' reset pin = output
   OUTA &= %11111111_11111111_11011111_11111111         ' set reset pin low   

PUB Z80ResetHigh
   DIRA |= %00000000_00000000_00100000_00000000         ' reset pin = output
   OUTA |= %00000000_00000000_00100000_00000000         ' set reset pin high

PUB Z80WaitForRdLow                                     ' keep clocking until the /rd pin goes low
   repeat while (INA & %00000000_00000000_00000100_00000000)
     Z80Clock(1)

PUB Z80WriteByteToRam(n)
    n &= %00000000_00000000_00000000_11111111           ' mask off all but lower 8 bits
    DIRA |= %00000000_00000000_00000110_11111111        ' set P0-P7 as outputs and also /RD and /WR as outputs
    OUTA &= %11111111_11111111_11111111_00000000        ' mask outa lower 8 bits to zero
    OUTA |= n                                           ' combine with the data byte
    OUTA &= %11111111_11111111_11111101_11111111        ' force /WR low
    OUTA |= %00000000_00000000_00000100_00000000        ' force /RD high
    'delay.pause1ms(1)                                   ' delay, make as short as possible
    OUTA |= %00000000_00000000_00000010_00000000        ' write high
    DIRA &= %11111111_11111111_11111001_00000000         'set /rd /wr and the data bits to inputs (highZ)

PUB Z80DataBusNOP                                       ' put zero (NOP) on data bus
    DIRA |= %00000000_00000000_00000000_11111111        ' set data bus to outputs
    OUTA &= %11111111_11111111_11111111_00000000        ' set data bus to zero                            

PUB ReceiveBlock(address)                               ' fetch 128 bytes from the Z80 at address
    WaitforIORQlow
    ' Z80 is now expecting a command. Prop pins 13 up are /iorq, /wr,/rd, clk, A0, reset
    ' status of prop pins 13 to 18 should be LHLLHH ie a read from port 1
    SendByte(129)                                       ' send command 129 - fetch block
    SendAddress(address)                                  ' send as a word
    repeat 128
      WaitforIORQlow                                      ' this should be a Z80 out command - /wr low
      PrintHexByte(INA)                                   ' send this byte to the terminal
      Z80Clock(2)                                         ' so iorq high again

PUB SendBlock(address,data)                             ' replace later with a block from, say SD card file
    WaitforIORQlow
    ' Z80 is now expecting a command. Prop pins 13 up are /iorq, /wr,/rd, clk, A0, reset
    ' status of prop pins 13 to 18 should be LHLLHH ie a read from port 1
    SendByte(128)                                       ' send command 128 - send block
    SendAddress(address)                                ' send as a word
    repeat 128
      WaitforIORQlow
      SendByte(data)                                    ' same data for the moment
      


PUB SendAddress(address)
    WaitforIORQlow                                       ' wait until ready for next command
   ' Z80 is now expecting the first byte of the address. Port is zero so A0 is low 
   ' status of prop pins 13 to 18 should be LHLLLH
    SendByte(address & %00000000_00000000_00000000_11111111) ' send low address byte
    WaitforIORQlow                                      ' ready for high byte of address
    SendByte(address >> 8)                              ' send high byte    

PUB NextInstruction                                     ' run until next instruction                                                      
    Z80Clock(2)                                         ' clock to clear last instruction                                                   
    WaitforRDlow      

PUB BusHiZ
     DIRA &= %11111111_11111111_11111111_00000000        ' set data pins to HiZ so doesn't clash with Z80

PUB SendByte(n)                                         ' put a byte on the Z80 bus, clock x2 then set bus HiZ
    n &= %00000000_00000000_00000000_11111111           ' mask off lower byte
    DIRA |= %00000000_00000000_00000000_11111111        ' set data pins as outputs
    OUTA &= %11111111_11111111_11111111_00000000        ' set data bus to zero
    OUTA |= n                                           ' merge with the byte to send
    Z80Clock(2)                                         ' two clocks to send
    BusHiZ                                              ' hand control of the bus back to the Z80

PUB WaitforIORQlow                                      ' clock until /iorq is low
    repeat while (INA & %00000000_00000000_00000001_00000000) ' clock until /iorq goes low
      Z80Clock(1)

PUB WaitforRDlow
    repeat while (INA & %00000000_00000000_00000100_00000000) ' clock until /rd goes low
      Z80Clock(1)        

PUB PrintStringCR(pstring)' sends pstring to the vga and serial port
     PrintString(pstring)                               ' print the string  
     crlf

PUB PrintString(pstring)' sends pstring to the serial port with no CRLF
    sio.str(0,pstring)                                 ' send to com port

PUB PrintChar(c) ' sends c to vga and serial port
      sio.tx(0,c)                                         ' send to com port
    
PUB CRLF
    PrintChar(13)                                        ' send to vga
    PrintChar(10)                                        ' send to vga

PUB PrintDecimal(number)                                ' print the decimal value, useful for debuggin
     PrintString(str.integerToDecimal(number, 10))

PUB PrintHexByte(number) ' print hex byte
    PrintString(str.integerToHexadecimal(number,2))   


{


  VT100.start(16)                                       ' start the vga driver VGA starts on pin 16
  VT100.startcursor                                     ' start the cursor
  VT100.color(%00001000_11111100)                       ' Blue_White  blue =3/4 power 10, white =full                                  
  VT100.cursorset(5)                                    ' cursor type 5
  VT100.clearscreen                                     ' clear the screen
  'vt100.str(string("Testing for SD card",13,10))        ' helpful message if card is out, better than just a blank screen                                  
  key.startx(27, 26, %100, 40)             'Start Keyboard Driver  pin,pin,num,repeatrate
  sio.AddPort(0, _receiverPin, _transmitterPin, -1, -1, 0, 0, _baudRateSpeed)
  'sio.AddPort(1, _port2receiverPin, _port2transmitterPin, -1, -1, 0, 0, 120000) ' shamcom works 108000 to 120000 (teraterm can't do 112800 even though it can select it)
  'sio.tx(1,65)    ' test port 2

  ifnot( fat.FATEngineStart(_SD_DO, _SD_CLK, _SD_DI, _SD_CS, _SD_WP, _SD_CD, _RTC_DAT, _RTC_CLK, -1))
'        com.COMEngineStart(_receiverPin, _transmitterPin, _baudRateSpeed)) 
'        adc.ADCEngineStart(_leftMicInputPin, _leftMicFeedBackPin, _rightMicInputPin, _rightMicFeedBackPin, 16_000)
'        dac.DACEngineStart(_leftChannelAudioPin, _rightChannelAudioPin, 0) )
    reboot

'  dac.leftChannelVolume(_leftChannelVolume)
'  dac.rightChannelVolume(_rightChannelVolume)

'  if(_VGA_Enabled)
'    ifnot( bmp.BMPEngineStart(_VGA_PinGroup, 1, _VGA_H_Res, _VGA_V_Res, @scopeHeap) and {
'         } (cognew(scope, @scopeStack) <> -1))
'      reboot

  sio.start             ' start the com port  note can't print anything to screen till sio.start unless only do to vga screen
  PrintStringCR(string("Testing for SD card"))  ' helpful message if card is out, better than just a blank screen
  fat.mountPartition(0)     ' mount the sd card
  VT100.clearscreen                                     ' if the sd card is not inserted, will leave the message on the screen as a reminder. If it is inserted, start with a blank screen
  PrintStringCR(string("*** KyeDOS SD card operating system v3.01 by Kwabena W. Agyeman & J. Moxham ***"))
  PrintString(string("Type Help for command listing"))
  crlf
  'DisplayBinaryFiles                    ' display all files ending in .BIN
  sio.rxflush(0)                                        ' in case try to send an xmodem transfer

  ifnot(fat.partitionMounted)
    printstring(string("Card Not Mounted "))

  Z80Clock                      ' start the Z80 cog clock
  'Z80TinyTestProgram            ' run a little test program
  'Z80TestHL
  Bootstrap                     ' load a bootstrap program into the Z80

  repeat

    result := shellLine(string(">"))
    'printstring(string("Running command: ")) ' commands run so quickly no need to print this
    'printstring(result)
    str.copy(result,@commandline)' save for later on if this is a command line with parameters eg "xmodem r filename.txt"
    crlf

    result := \shellCommands(result)
    printstring(result)
    crlf

    if(fat.partitionError)
      crlf

PRI shellCommands(stringPointer)

  stringPointer := str.tokenizeString(stringPointer)

  programClear(stringPointer)
  programEcho(stringPointer)

  programReboot(stringPointer)
  programHelp(stringPointer)

  'programMount(stringPointer)
  'programUnmount(stringPointer)

  programFreeSpace(stringPointer)
  programUsedSpace(stringPointer)

  programList(stringPointer)
  programDIR(stringPointer)
  programConcatenate(stringPointer)

  programChangeDirectory(stringPointer)
  programChangeAttributes(stringPointer)

  programMove(stringPointer)
  programRemove(stringPointer)
  programEra(stringPointer)  ' same as programremove

  programMakeFile(stringPointer)
  programMakeDirectory(stringPointer)

  programCopy(stringPointer)
  programDifference(stringPointer)

  programBoot(stringPointer)
  programFormat(stringPointer)

'  programDate(stringPointer)                                      
'  programTime(stringPointer)

'  programRamSet(stringPointer)
'  programRamGet(stringPointer)

'  programSQW(stringPointer)
'  programOUT(stringPointer)

'  programPlayWavFile(stringPointer)
'  programTapeWavFile(stringPointer)

  programDump(stringPointer)
  'programTest(stringPointer)

  programCommand(StringPointer) ' is it a command ie a .exe file?

  abort string("Command Not Found! Try Help")
  crlf
{
PRI shellLine(prompt)

  printstring(prompt)

  repeat
    result := com.receivedByte
    str.buildString(result)
  while((result <> carriage_return) and (result <> carriage_return) and (result <> com#null))

  return str.trimString(str.builtString(true))
}

PRI shellLine(prompt)

  PrintString(prompt)
  repeat

      result:=0
      repeat
        if key.gotkey <>0        ' scan both the local keyboard and the com port
          result:=key.getkey
        if sio.rxavail(0)
          result :=sio.rx(0)  
      until result <>0           ' until something comes in either port
    
    if result==200
      result :=8                ' backspace keycode is 200 but convert to 8
    PrintChar(result)             ' send to vga and to com port
    str.buildString(result)     ' build an input string
  while((result <> 13) and (result <> 10) and (result <> 0))

  return str.trimString(str.builtstring(true))

{
PRI shellDecision(prompt, characterToFind, otherCharacterToFind)

  printstring(prompt)

  repeat

    prompt := com.receivedByte
    result or= ((prompt == characterToFind) or (prompt == otherCharacterToFind))

    if(prompt == com#backspace)
      return false

  while((prompt <> carriage_return) and (prompt <> carriage_return) and (prompt <> com#null))
}
PRI shellDecision(prompt, characterToFind, otherCharacterToFind)

  PrintString(prompt~)  

  repeat
    result := key.getkey
    prompt or= ((result == characterToFind) or (result == otherCharacterToFind))
  while((result <> 13) and (result <> 10) and (result <> 0))  

  return prompt  

PRI programClear(stringPointer)

  ifnot(str.stringCompareCI(string("clear"), stringPointer))
    delay.pause1ms(500)
    'rtc.pauseForMilliseconds(500)
    printstring(string(_clearScreenCharacter, _homeCursorCharacter))
    abort false

PRI programEcho(stringPointer)

  ifnot(str.stringCompareCI(string("echo"), stringPointer))
    printstring(str.trimString(stringPointer += 5))
    crlf
    abort false

PRI programReboot(stringPointer)

  ifnot(str.stringCompareCI(string("reboot"), stringPointer))
    printstring(string("Rebooting "))

    repeat 3
      'rtc.pauseForMilliseconds(500)
      delay.pause1ms(500)  
      printstring(string(". "))

    'rtc.pauseForMilliseconds(500)
    delay.pause1ms(500)  
    reboot

PRI programHelp(stringPointer)

  ifnot(str.stringCompareCI(string("help"), stringPointer))
    printstring(@shellProgramHelpStrings)
    abort false

DAT shellProgramHelpStrings

  byte 13,10
  byte "Command Listing"
  byte 13,10
  byte "<clear>                        - Clear the screen."
  byte 13,10
  byte "<echo> <string>                - Echo the string argument."
  byte 13,10
  byte "<reboot>                       - Reboot the propeller chip."
  byte 13,10
  byte "<help>                         - Show the command listing."
  byte 13,10
  'byte "<mount> <partition>            - Mount the file system."
  'byte 13,10
  'byte "<unmount>                      - Unmount the file system."
  'byte 13,10
  byte "<df> <fast>                    - Free sector count. F=Fast. 1 Sector = 512 Bytes"
  byte 13,10
  byte "<du> <fast>                    - Used sector count. F=Fast. 1 Sector = 512 Bytes"
  byte 13,10
  byte "<ls> <all>                     - List contents of the working directory. A=All."
  byte 13,10
  byte "<dir>                          - List contents. Includes wildcards eg dir *.exe"
  byte 13,10
  byte "<cat> <file> <hex>             - Print the ASCII interpreted contents of a file."
  byte 13,10
  byte "<cd> <directory>               - Change directory."
  byte 13,10
  byte "<chmod> <entry> <attributes>   - Change Attrib: R=R/O, H=Hidden, S=Sys, A=Arch"
  byte 13,10
  byte "<mv> <oldPath> <newPath>       - Move a file or directory."
  byte 13,10
  byte "<rm> <path> or <era>           - Remove a file or directory."
  byte 13,10
  byte "<mkfile> <name>                - Make a new file."
  byte 13,10
  byte "<mkdir> <name>                 - Make a new directory."
  byte 13,10
  byte "<cp> <oldPath> <newPath>       - Copy a file."
  byte 13
  byte 10
  byte "<diff> <path> <path>           - Print differences between two files."
  byte 13,10
  byte "<boot> <file>                  - Reboot the propeller chip from a file."
  byte 13,10
  byte "<filename>                     - Reboot and run filename.exe"
  byte 13,10
  byte "<format> <partition>           - Format and mount the file system."
  byte 13,10
  'byte "<date>                         - Display the current date and time."
  'byte 13,10
  'byte "<time>                         - Change the current date and time."
  'byte 13,10
'  byte "<set> <byteNumber> <byteValue> - Set the value of a SRAM byte on the RTClock."
'  byte 13,10
'  byte "<get> <byteNumber>             - Get the value of a SRAM byte on the RTClock."
'  byte 13,10
'  byte "<sqw> <frequencyNumber>        - Set the frequecny of the SWQ / OUT pin. (0=1Hz, 1=4096Hz, 2=8192Hz, 3=32768Hz)."
'  byte 13,10
'  byte "<out> <digitalState>           - Set the state of the SWQ / OUT pin. (0=low, 1=high)."
'  byte 13,10
'  byte "<play> <file>                  - Play a WAV file."
'  byte 13,10
'  byte "<tape> <file>                  - Tape a WAV file."
'  byte 13,10
  byte "<dump>                         - Dump the ROM font."
  byte 13,10
  'byte "<test>                         - Speed test the file system."
  'byte 13,10,
  byte 0

'PRI programMount(stringPointer)
'
'  ifnot(str.stringCompareCI(string("mount"), stringPointer))
'    if(fat.mountPartition(stringPointer := str.decimalToInteger(str.tokenizeString(0))))
'      printstringcr(string("Please unmount the disk next time"))
'    taf.mountPartition(stringPointer)
'
'    printstring(fat.partitionWriteProtected & string("Write Protected "))
'    printstring(string("Disk 0x"))
'    printstring(str.integerToHexadecimal(fat.partitionDiskSignature, 8))
'    printstring(string(" with Volume ID 0x"))
'    printstring(str.integerToHexadecimal(fat.partitionVolumeIdentification, 8))
'    printstring(string(" a.k.a "))
'    printstring(str.trimString(fat.partitionVolumeLabel))
'    printstringcr(string(" ready"))
'
'    abort 0
'
'PRI programUnmount(stringPointer)
'
'  ifnot(str.stringCompareCI(string("unmount"), stringPointer))
'    taf.unmountPartition
'    fat.unmountPartition
'    printstringcr(string("Disk ready for removal"))
'
'    abort 0

PRI programFreeSpace(stringPointer)

  ifnot(str.stringCompareCI(string("df"), stringPointer))

    printstring(string("The ", quote))
    printstring(str.trimString(fat.partitionFileSystemType))
    printstring(string(quote, " ", quote))
    printstring(str.trimString(fat.partitionVolumeLabel))
    printstring(string(quote, " parition has "))
    printstring(1 + str.integerToDecimal(fat.partitionCountOfClusters, 10))
    printstring(string(" clusters and "))
    printstring(1 + str.integerToDecimal(fat.partitionDataSectors, 10))
    printstring(string(" total sectors, where each cluster is "))
    printstring(1 + str.integerToDecimal(fat.partitionSectorsPerCluster, 3))
    printstring(string(" sectors large and each sector is "))
    printstring(1 + str.integerToDecimal(fat.partitionBytesPerSector, 3))
    printstringcr(string(" bytes"))
    printstring(string("And there are... please wait... "))
    printstring(1 + str.integerToDecimal(fat.partitionFreeSectorCount(byte[str.tokenizeString(0)]), 10))
    printstringcr(string(" Free Sectors"))

    abort 0

PRI programUsedSpace(stringPointer)

  ifnot(str.stringCompareCI(string("du"), stringPointer))

    printstring(string("The ", quote))
    printstring(str.trimString(fat.partitionFileSystemType))
    printstring(string(quote, " ", quote))
    printstring(str.trimString(fat.partitionVolumeLabel))
    printstring(string(quote, " parition has "))
    printstring(1 + str.integerToDecimal(fat.partitionCountOfClusters, 10))
    printstring(string(" clusters and "))
    printstring(1 + str.integerToDecimal(fat.partitionDataSectors, 10))
    printstring(string(" total sectors where each cluster is "))
    printstring(1 + str.integerToDecimal(fat.partitionSectorsPerCluster, 3))
    printstring(string(" sectors large and each sector is "))
    printstring(1 + str.integerToDecimal(fat.partitionBytesPerSector, 3))
    printstringcr(string(" bytes"))
    printstring(string("And there are... please wait... "))
    printstring(1 + str.integerToDecimal(fat.partitionUsedSectorCount(byte[str.tokenizeString(0)]), 10))
    printstringcr(string(" Used Sectors"))

    abort 0

PRI programList(stringPointer) : counter | buffer, flag

  ifnot(str.stringCompareCI(string("ls"), stringPointer))
    stringPointer := str.tokenizeString(0)
    flag := (str.findCharacter(stringPointer, "A") or str.findCharacter(stringPointer, "a"))
    fat.listEntries("W")

    repeat while(buffer := fat.listEntries("N"))
      printstring(buffer)
      ifnot(flag)

        if(++counter // 8)
          PrintChar(horizontal_tab)
        else
          crlf

      else

        if(fat.listIsReadOnly)
          printstring(string(" [ R"))
        else
          printstring(string(" [ _"))

        if(fat.listIsHidden)
          PrintChar("H")
        else
          PrintChar("_")

        if(fat.listIsSystem)
          PrintChar("S")
        else
          PrintChar("_")

        if(fat.listIsDirectory)
          PrintChar("D")
        else
          PrintChar("_")

        if(fat.listIsArchive)
          PrintChar("A")
        else
          PrintChar("_")

        printstring(string(" ] [ "))
        printstring(1 + str.integerToDecimal(fat.listSize, 10))

        printstring(string(" ] [ Accessed "))
        printstring(1 + str.integerToDecimal(fat.listAccessMonth, 2))
        PrintChar("/")
        printstring(1 + str.integerToDecimal(fat.listAccessDay, 2))
        PrintChar("/")
        printstring(1 + str.integerToDecimal(fat.listAccessYear, 4))

        printstring(string(" ] [ Created "))
        printstring(1 + str.integerToDecimal(fat.listCreationHours, 2))
        PrintChar(":")
        printstring(1 + str.integerToDecimal(fat.listCreationMinutes, 2))
        PrintChar(":")
        printstring(1 + str.integerToDecimal(fat.listCreationSeconds, 2))

        PrintChar(" ")
        printstring(1 + str.integerToDecimal(fat.listCreationMonth, 2))
        PrintChar("/")
        printstring(1 + str.integerToDecimal(fat.listCreationDay, 2))
        PrintChar("/")
        printstring(1 + str.integerToDecimal(fat.listCreationYear, 4))

        printstring(string(" ] [ Modified "))
        printstring(1 + str.integerToDecimal(fat.listModificationHours, 2))
        PrintChar(":")
        printstring(1 + str.integerToDecimal(fat.listModificationMinutes, 2))
        PrintChar(":")
        printstring(1 + str.integerToDecimal(fat.listModificationSeconds, 2))

        PrintChar(" ")
        printstring(1 + str.integerToDecimal(fat.listModificationMonth, 2))
        PrintChar("/")
        printstring(1 + str.integerToDecimal(fat.listModificationDay, 2))
        PrintChar("/")
        printstring(1 + str.integerToDecimal(fat.listModificationYear, 4))

        printstringcr(string(" ]"))
    if((counter // 8) and counter) 
      crlf

    abort 0

PRI programConcatenate(stringPointer) : counter | buffer, flag

  ifnot(str.stringCompareCI(string("cat"), stringPointer))
    fat.openFile(str.tokenizeString(0), "R")
    stringPointer := str.tokenizeString(0)
    flag := (str.findCharacter(stringPointer, "H") or str.findCharacter(stringPointer, "h"))

    repeat until(fat.fileSize == fat.fileTell)
      if(flag)
        PrintChar("$")
        printstring(str.integerToHexadecimal(fat.readByte, 2))

        if(++counter // 16)
          PrintChar(horizontal_tab)
        else
          crlf

      else
        PrintChar(fat.readByte)

    if((counter // 8) and counter)
      crlf

    fat.closeFile
    abort 0

PRI programChangeDirectory(stringPointer)

  ifnot(str.stringCompareCI(string("cd"), stringPointer))
    stringPointer := str.tokenizeString(0)
    fat.changeDirectory(stringPointer)
    taf.changeDirectory(stringPointer)
    abort 0

PRI programChangeAttributes(stringPointer)

  ifnot(str.stringCompareCI(string("chmod"), stringPointer))
    fat.changeAttributes(str.tokenizeString(0), str.tokenizeString(0))
    abort 0

PRI programMove(stringPointer)

  ifnot(str.stringCompareCI(string("mv"), stringPointer))
    fat.moveEntry(str.tokenizeString(0), str.tokenizeString(0))
    abort 0

PRI programRemove(stringPointer)

  ifnot(str.stringCompareCI(string("rm"), stringPointer)) 
    fat.deleteEntry(str.tokenizeString(0))
    abort 0

PRI programERA(stringPointer) ' same as ProgramRemove

  ifnot(str.stringCompareCI(string("era"), stringPointer)) 
    fat.deleteEntry(str.tokenizeString(0))
    abort 0

PRI programMakeFile(stringPointer)

  ifnot(str.stringCompareCI(string("mkfile"), stringPointer))
    fat.newFile(str.tokenizeString(0))
    abort 0

PRI programMakeDirectory(stringPointer)

  ifnot(str.stringCompareCI(string("mkdir"), stringPointer))
    fat.newDirectory(str.tokenizeString(0))
    abort 0

PRI programCopy(stringPointer) | sector[128]

  ifnot(str.stringCompareCI(string("cp"), stringPointer))
    fat.openfile(str.tokenizeString(0), "R")
    stringPointer := str.tokenizeString(0)
    taf.newFile(stringPointer)
    taf.openfile(stringPointer, "W")

    repeat
      result := fat.readData(@sector, 512)
      taf.writeData(@sector, result)
    while(result == 512)

    fat.closeFile
    taf.closeFile
    abort 0

PRI programDifference(stringPointer) | byte0, byte1

  ifnot(str.stringCompareCI(string("diff"), stringPointer))
    fat.openfile(str.tokenizeString(0), "R")
    taf.openfile(str.tokenizeString(0), "R")

    repeat until(fat.fileSize == fat.fileTell)
      byte0 := fat.readByte
      byte1 := taf.readByte

      if(byte0 <> byte1)
        printstring(string("Difference: $"))
        printstring(1 + str.integerToHexadecimal(byte0, 2))
        printstring(string(" <> $"))
        printstring(1 + str.integerToHexadecimal(byte1, 2))
        printstring(string(" at $"))
        printstringcr(1 + str.integerToHexadecimal(result, 8))

      result += 1

    fat.closeFile
    taf.closeFile
    abort 0

PRI programBoot(stringPointer)

  ifnot(str.stringCompareCI(string("boot"), stringPointer))
      stringPointer := str.tokenizeString(0)
    'printstring(string("Are you sure you want to reboot from "))
    'printstring(stringPointer)
    'if(shellDecision(string(" - [y]es or [n]o? : "), "Y", "y"))
      fat.bootPartition(stringPointer)

    abort 0

PRI programFormat(stringPointer)

  ifnot(str.stringCompareCI(string("format"), stringPointer))
    stringPointer := str.tokenizeString(0)

    printstring(string("Are you sure you want to format ", quote))
    printstring(fat.partitionVolumeLabel)
    if(shellDecision(string(quote," All data will be erased! - [y]es or [n]o? : "), "Y", "y"))
      if(shellDecision(string("Are you absolutely sure? [y]es or [n]o : "), "Y", "y"))
        printstringcr(string("Formatting Partition..."))
        fat.formatPartition(str.decimalToInteger(str.tokenizeString(0)))

    abort 0
    
{
PRI programDate(stringPointer)

  ifnot(str.stringCompareCI(string("date"), stringPointer))

    ifnot(rtc.readTime)
      abort string("Operation Failed")
      crlf

    printstring(lookup(rtc.clockDay: string("Sunday"), {
                                          } string("Monday"), {
                                          } string("Tuesday"),  {
                                          } string("Wednesday"), {
                                          } string("Thursday"), {
                                          } string("Friday"), {
                                          } string("Saturday")))
    printstring(string(", "))
    printstring(lookup(rtc.clockMonth: string("January"), {
                                            } string("Feburary"), {
                                            } string("March"), {
                                            } string("April"), {
                                            } string("May"), {
                                            } string("June"), {
                                            } string("July"),  {
                                            } string("August"), {
                                            } string("September"), {
                                            } string("October"), {
                                            } string("November"), {
                                            } string("December")))

    PrintChar(" ")
    printstring(1 + str.integerToDecimal(rtc.clockDate, 2))
    printstring(string(", "))
    printstring(1 + str.integerToDecimal(rtc.clockYear, 4))
    PrintChar(" ")
    printstring(1 + str.integerToDecimal(rtc.clockMeridiemHour, 2))
    PrintChar(":")
    printstring(1 + str.integerToDecimal(rtc.clockMinute, 2))
    PrintChar(":")
    printstring(1 + str.integerToDecimal(rtc.clockSecond, 2))
    PrintChar(" ")

    if(rtc.clockMeridiemTime)
      printstringcr(string("PM"))
    else
      printstringcr(string("AM"))

    abort false

PRI programTime(stringPointer) | time[6]

  ifnot(str.stringCompareCI(string("time"), stringPointer))
    printstringcr(string("Please enter the exact time,"))

    time := str.decimalToInteger(shellLine(string("Year (2000 - 2127): ")))
    time[1] := str.decimalToInteger(shellLine(string("Month (1 - 12): ")))
    time[2] := str.decimalToInteger(shellLine(string("Date (1 - 31): ")))
    time[3] := str.decimalToInteger(shellLine(string("Day (1 - 7): ")))
    time[4] := str.decimalToInteger(shellLine(string("Hours (0 - 23): ")))
    time[5] := str.decimalToInteger(shellLine(string("Minutes (0 - 59): ")))
    ifnot(rtc.writeTime(str.decimalToInteger(shellLine(string("Seconds (0 - 59): "))), {
                       } time[5], {
                       } time[4], {
                       } time[3], {
                       } time[2], {
                       } time[1], {
                       } time))
      abort string("Operation Failed")
      crlf

    crlf
    programDate(string("date"))
    printstringcr(string("Operation Successful"))
    abort false
    

PRI programRamSet(stringPointer)

  ifnot(str.stringCompareCI(string("set"), stringPointer))
    result := rtc.writeSRAM(str.decimalToInteger(str.tokenizeString(0)), str.decimalToInteger(str.tokenizeString(0)))

    stringPointer := string("Failure!", _newLineCharacter)
    if(result)
      stringPointer := string("Success!", _newLineCharacter)

    printstring(stringPointer)
    abort false

PRI programRamGet(stringPointer)

  ifnot(str.stringCompareCI(string("get"), stringPointer))
    printstring(1 + str.integerToDecimal(rtc.readSRAM(str.decimalToInteger(str.tokenizeString(0))), 3))
    PrintChar(_newLineCharacter)
    abort false
}
{
PRI programSQW(stringPointer)

  ifnot(str.stringCompareCI(string("sqw"), stringPointer))
    result := rtc.clockSquareWaveOut(str.decimalToInteger(str.tokenizeString(0)) & 1, 1)

    stringPointer := string("Failure!", _newLineCharacter)
    if(result)
      stringPointer := string("Success!", _newLineCharacter)

    printstring(stringPointer)
    abort false

PRI programOUT(stringPointer)

  ifnot(str.stringCompareCI(string("out"), stringPointer))
    result := rtc.clockSquareWaveOut(0, ((str.decimalToInteger(str.tokenizeString(0)) & 1) << 1))

    stringPointer := string("Failure!", _newLineCharacter)
    if(result)
      stringPointer := string("Success!", _newLineCharacter)

    printstring(stringPointer)
    abort false
}    
{
PRI programPlayWavFile(stringPointer) | numberOfChannels, sampleRate ' Google "WAV format" for info about this function...

  ifnot(str.stringCompareCI(string("play"), stringPointer))
    printstring(string("Please wait...", _newLineCharacter))
    stringPointer := fat.openFile(str.tokenizeString(0), "R")

    if(fat.readLong <> $46_46_49_52)
      abort string("Not RIFF file", _newLineCharacter)

    if((fat.readlong + 8) <> fat.fileSize)
      abort string("Chunk size incorrect", _newLineCharacter)

    if(fat.readLong <> $45_56_41_57)
      abort string("Not WAVE file", _newLineCharacter)

    if(fat.readLong <> $20_74_6D_66)
      abort string(quote, "fmt ", quote, " missing",  _newLineCharacter)

    if(fat.readLong <> 16)
      abort string("Subchunk 1 size incorrect", _newLineCharacter)

    if(fat.readShort <> 1)
      abort string("Not LPCM file", _newLinecharacter)

    dac.numberOfChannelsSetup(numberOfChannels := fat.readShort)
    dac.sampleRateSetup(sampleRate := fat.readLong)

    fat.fileSeek(34)
    dac.bitsPerSampleSetup(result := fat.readShort)
    dac.sampleSignSetup(result == 16)

    fat.fileSeek(28)

    if(fat.readLong <> (sampleRate * numberOfChannels * result / 8))
      abort string("Byte rate incorrect", _newLineCharacter)

    if(fat.readShort <> (numberOfChannels * result / 8))
      abort string("Block align incorrect", _newLineCharacter)

    fat.fileSeek(36)

    if(fat.readLong <> $61_74_61_64)
      abort string(quote, "data", quote, " missing",  _newLineCharacter)

    result := fat.readlong
    if((result + constant(36 + 8)) <> fat.fileSize)
      abort string("Subchunk 2 size incorrect", _newLineCharacter)

    printstring(string("Now playing "))
    printstring(stringPointer)
    printstring(string(_newLineCharacter, "Press enter to quit ", _newLineCharacter))

    dac.startPlayer
    scopeSwtich := true

    repeat ((result + 511) / 512)
      fat.readData(dac.transferData, 512)

      if(com.receivedNumber)
        result := com.receivedByte
        if((result == carriage_return) or (result == carriage_return) or (result == com#null))
          printstring(string("Quiting", _newLineCharacter))
          quit

    scopeSwtich := false
    dac.stopPlayer
    dac.clearData
    fat.closeFile
    abort 0
}    
{
PRI programTapeWavFile(stringPointer) ' Google "WAV format" for info about this function...

  ifnot(str.stringCompareCI(string("tape"), stringPointer))
    printstring(string("Please wait...", _newLineCharacter))

    stringPointer := str.tokenizeString(0)
    fat.newFile(stringPointer)
    stringPointer := fat.openFile(stringPointer, "W")

    printstring(string("Now taping "))
    printstring(stringPointer)
    printstring(string(_newLineCharacter, "Press enter to save ", _newLineCharacter))

    fat.writeString(string("RIFF____WAVEfmt "))
    fat.writeLong(16)
    fat.writeShort(1)
    fat.writeShort(1)
    fat.writeLong(16_000)
    fat.writeLong(16_000)
    fat.writeShort(1)
    fat.writeShort(8)
    fat.writeString(string("data____"))

    adc.sampleRateSetup(16_000)
    adc.numberOfChannelsSetup(1)
    adc.bitsPerSampleSetup(8)
    adc.sampleSignSetup(0)

    adc.clearData
    repeat while(result < (posx - constant(36 + 8)))
      fat.writeData(adc.transferData, 512)
      result += 512

      if(com.receivedNumber)
        stringPointer := com.receivedByte
        if((stringPointer == carriage_return) or (stringPointer == carriage_return) or (stringPointer == com#null))
          printstring(string("Saving", _newLineCharacter))
          quit

    fat.fileSeek(40)
    fat.writeLong(result)
    fat.fileSeek(4)
    fat.writeLong(result + 36)

    fat.closeFile
    abort 0
}
PRI programDump(stringPointer) | outsideCounter, insideCounter

  ifnot(str.stringCompareCI(string("dump"), stringPointer))
    fat.openFile(fat.newFile(string("P8X32A.TXT")), "W")

    fat.writeString(string("P8X32A Character Set", carriage_return, line_feed, carriage_return, line_feed))
    repeat outsideCounter from 0 to 255

      printstring(string("Dumping character #"))
      printstringcr(1 + str.integerToDecimal(outsideCounter, 3))

      fat.writeString(string("Character #"))
      fat.writeString(1 + str.integerToDecimal(outsideCounter, 3))
      fat.writeString(string(carriage_return, carriage_return, carriage_return, carriage_return))

      repeat result from 0 to 31
        repeat insideCounter from (outsideCounter & 1) to 31 step 2

          if(long[32_768][result + ((outsideCounter >> 1) << 5)] & (|<insideCounter))
            fat.writeByte(".")
          else
            fat.writeByte(" ")

        fat.writeString(string(carriage_return, carriage_return))
      fat.writeString(string(carriage_return, carriage_return))

    fat.closeFile
    abort 0
{
PRI programTest(stringPointer) | counter, buffer

  ifnot(str.stringCompareCI(string("test"), stringPointer))
    fat.newFile(stringPointer := string("P8X32A.ROM"))

    buffer := startTest(stringPointer, "W", string("Byte write test...   "))
    repeat counter from 32_768 to 65_535 step 1
      fat.writeByte(byte[counter])
    stopTest("W", 32_768, (cnt - buffer))

    buffer := startTest(stringPointer, "R", string("Byte read test...    "))
    repeat counter from 32_768 to 65_535 step 1
      if(fat.readByte <> byte[counter])
        fat.closeFile
        abort string("Byte read back failed")
        crlf
    stopTest("R", 32_768, (cnt - buffer))

    buffer := startTest(stringPointer, "W", string("Word write test...   "))
    repeat counter from 32_768 to 65_535 step 2
      fat.writeShort(word[counter])
    stopTest("W", 32_768, (cnt - buffer))

    buffer := startTest(stringPointer, "R", string("Word read test...    "))
    repeat counter from 32_768 to 65_535 step 2
      if(fat.readShort <> word[counter])
        fat.closeFile
        abort string("Word read back failed")
        crlf
    stopTest("R", 32_768, (cnt - buffer))

    buffer := startTest(stringPointer, "W", string("Long write test...   "))
    repeat counter from 32_768 to 65_535 step 4
      fat.writeLong(long[counter])
    stopTest("W", 32_768, (cnt - buffer))

    buffer := startTest(stringPointer, "R", string("Long read test...    "))
    repeat counter from 32_768 to 65_535 step 4
      if(fat.readLong <> long[counter])
        fat.closeFile
        abort string("Long read back failed")
        crlf
    stopTest("R", 32_768, (cnt - buffer))

    buffer := startTest(stringPointer, "W", string("Sector write test... "))
    repeat counter from 32_768 to 65_535 step 512
      fat.writeData(counter, 512)
    stopTest("W", 32_768, (cnt - buffer))

    buffer := startTest(stringPointer, "R", string("Sector read test...  "))
    repeat counter from 32_768 to 65_535 step 512
      fat.readData(counter, 512)
    stopTest("R", 32_768, (cnt - buffer))

    buffer := startTest(stringPointer, "W", string("Memory write test... "))
    repeat counter from 32_768 to 65_535 step 32_768
       fat.writeData(counter, 32_768)
    stopTest("W", 32_768, (cnt - buffer))

    buffer := startTest(stringPointer, "R", string("Memory read test...  "))
    repeat counter from 32_768 to 65_535 step 32_768
      fat.readData(counter, 32_768)
    stopTest("R", 32_768, (cnt - buffer))

    fat.newFile(fat.deleteEntry(stringPointer))
    buffer := startTest(stringPointer, "A", string("Block append test... "))
    fat.writeData(32_768, 32_768)
    stopTest("W", 32_768, (cnt - buffer))

    buffer := startTest(stringPointer, "R", string("Reverse seek test... "))
    repeat counter from 32_767 to 0 step 512
      fat.fileSeek(counter)
    stopTest("R", 32_768, (cnt - buffer))

    fat.deleteEntry(stringPointer)
    abort 0

PRI startTest(path, mode, message)

  fat.openFile(path, mode)
  printstring(message)
  return cnt

PRI stopTest(mode, count, cycles)

  if(mode == "W")
    printstring(string("Writ "))
  else
    printstring(string("Read "))

  printstring(1 + str.integerToDecimal((count >> 10), 3))
  printstring(string(" KBs at "))
  printstring(1 + str.integerToDecimal((multiplyDivide(count, cycles) >> 10), 3))
  printstringcr(string(" KBs per second"))
  fat.closeFile
}
PRI multiplyDivide(dividen, divisor) | productHigh, productLow

  productHigh := clkfreq ** dividen
  productLow := clkfreq * dividen
  if(productHigh => divisor)
    return posx

  repeat 32

    dividen := (productHigh < 0)
    productHigh := ((productHigh << 1) + (productLow >> 31))
    productLow <<= 1
    result <<= 1

    if((productHigh => divisor) or dividen)
      productHigh -= divisor
      result += 1
{
PRI scope | displayPointer, currentDisplay

  bmp.displayColor(0, _VGA_BG_Color)
  bmp.displayColor(1, _VGA_FG_Color)
  displayPointer := constant((_VGA_H_Res * _VGA_V_Res) / 32)
  repeat

    bmp.displayWait(1)
    bmp.displayPointer(@scopeHeap[displayPointer & currentDisplay])
    not currentDisplay
    bmp.displayClear(0, @scopeHeap[displayPointer & currentDisplay])

    repeat result from 0 to constant(_VGA_H_Res - 1)

      if(scopeSwtich)

        bmp.plotPixel( 1, {
                     } result, {
                     } (constant(_VGA_V_Res / 2) + (dac.leftChannelOutput ~> constant(16 - (>|_VGA_V_Res)))), {
                     } @scopeHeap[displayPointer & currentDisplay])
      else

        bmp.plotPixel( 1, {
                     } result, {
                     } (constant(_VGA_V_Res / 2) + (adc.leftChannelInput ~> constant(16 - (>|_VGA_V_Res)))), {
                     } @scopeHeap[displayPointer & currentDisplay])
}                     

PUB PrintStringCR(pstring)' sends pstring to the vga and serial port
     PrintString(pstring)                               ' print the string  
     crlf

PUB PrintString(pstring)' sends pstring to the vga and serial port with no CRLF
    if (_VGA_Enabled)  
      vt100.str(pstring)                                 ' send to vga screen
    if (_Serial_Enabled)   
    sio.str(0,pstring)                                 ' send to com port

PUB PrintChar(c) ' sends c to vga and serial port
    if (_VGA_Enabled)
      vt100.out(c)                                        ' send to vga
    if (_Serial_Enabled)
      sio.tx(0,c)                                         ' send to com port
    
PUB CRLF
    PrintChar(13)                                        ' send to vga
    PrintChar(10)                                        ' send to vga

PUB PrintDecimal(number)                                ' print the decimal value, useful for debuggin
     PrintString(str.integerToDecimal(number, 10))   

PRI DisplayBinaryFiles | buffer
  PrintString(string("Use the BOOT command to run the following binary Spin programs/emulations:"))
  crlf
  fat.listEntries("W")   
  repeat while(buffer := fat.listEntries("N"))   
      buffer := str.trimstring(buffer)   
      if str.endswithstring(buffer,string(".BIN")) == -1
        PrintString(buffer)
         crlf
  crlf

PRI ProgramDir(StringPointer) | buffer,counter,match
  ifnot(str.stringCompareCI(string("dir"), stringPointer))
    stringPointer := str.tokenizeString(0)              ' get the *.exe bit
    str.copy(stringpointer,@lineoftext1)                ' move to lineoftext1
    str.mid(@lineoftext1,@lineoftext2,str.instr(@lineoftext1,"."),4) ' get string from . to end = extension
    counter:=0
    crlf
    fat.listEntries("W")   
    repeat while(buffer := fat.listEntries("N"))
      str.copy(buffer,@LineOfText1)                ' copy filename to LineOfText
      str.trimstring(@lineoftext1)
      str.mid(@lineoftext1,@lineoftext3,str.instr(@lineoftext1,"."),4) ' get string from . to end = extension     
      match := false                                    ' setup for below
      if str.len(@lineoftext2) == 0
        match := true                                   ' no extension so always print it
      else
        ifnot str.stringcompareci(@lineoftext2,@lineoftext3) ' reverse logic ?!
          match := true                                 ' found a match
      if match == true
        PrintString(@lineoftext1)                       ' print file name
        counter++                                       ' add one to counter
        repeat 13-str.len(@lineoftext1)                 ' add spaces so columns line up
            printchar(" ")                              ' print space
      if counter == 6                                   ' new line
        counter :=0
        crlf 
    crlf
    abort 0                                             ' so knows a command was run

PRI ProgramCommand(StringPointer) | buffer,i ' look for .exe files (ie .binary files renamed to .exe) and if found, run the program
    fat.listEntries("W")
    str.copy(stringpointer,@lineoftext2)                ' copy the command to lineoftext2
    str.stringConcatenate(@lineoftext2, string(".exe")) ' add extension for executable files to lineoftext2
     repeat while(buffer := fat.listEntries("N"))
       str.copy(buffer,@LineOfText1)                    ' copy filename to LineOfText1
       str.trimstring(@lineoftext1)                     ' remove any leading/trailing spaces  
       ifnot str.stringcompareci(@lineoftext1,@lineoftext2)' does it match? (ignore case)
         str.copy(buffer,@lineoftext3)                  ' can't use "buffer" as it gets overwritten
         'printstringcr(string("Saving command line in COMMAND.TXT and running program"))
         'printstring(string("Command line: "))
         'printstring(@commandline)
          ' above commented out as 1) about to blank the screen, 2) debugged and 3) if running xmodem don't want any further data on the serial lines
          ' possibly could put this output back if use the vga screen only.                   
         result := \fat.newFile(string("COMMAND.TXT")) ' All aborts traps must have "variable := \function" setup.
         if(fat.partitionError == fat#Entry_Already_Exist)  ' file already exists
            result := \fat.deleteEntry(string("COMMAND.TXT"))    ' delete the old file
            result := \fat.newfile(string("COMMAND.TXT")) ' recreate the file
         result := \fat.openFile(string("COMMAND.TXT"), "W") ' create file and open file
         result := \fat.writestring(@commandline)       ' store the command line as it was typed
         result := \fat.closefile                       ' close the file
         result := \fat.bootpartition(@lineoftext3)     ' reboot


' ***************************************************************************************
      'printstring(@lineoftext1)                         ' testing string functions
      'str.left(@lineoftext1,@lineoftext2,3)  ' find leftmost characters in a string
      'printstring(@lineoftext2)
      'str.trimstring(@lineoftext1)
      'printstring(@lineoftext1)
      'str.trimstring(@lineoftext1)
      'str.mid(@lineoftext1,@lineoftext2,4,2) ' find middle characters in a string
      'printstring(@lineoftext2)
      'printstring(str.integertodecimal(str.len(@lineoftext2),5))
      'str.copy(string("hello"),@lineoftext1) ' copy fixed value into a string
      'printstring(@lineoftext1)
      'str.str(@lineoftext1,55) ' decimal to string with no leading zeros
      'printstring(@lineoftext1)
      'str.stringfill(@lineoftext1,10,65)
      'printstring(@lineoftext1)
      'str.str(@lineoftext1,1234)' convert to string
      'str.str(@lineoftext1,str.decimaltointeger(@lineoftext1)) ' convert back
      'printstring(@lineoftext1)
      ' str.left(@lineoftext1,@lineoftext2,str.len(@lineoftext1)-4) ' remove extension
      ' printstring(@lineoftext2) ' print it out


' **************** Z80 Clock **********************************
PUB Z80Clock                    ' start the Z80 clock
      cognew(@zToggle, 0) 'Launch new cog
DAT
{Toggle P15}
        org 0                   'Begin at Cog RAM addr 0
zToggle mov dira, Pin           'Set Pin to output
        mov Time, cnt           'Calculate delay time
        add Time, #9            'Set minimum delay here
:zloop  waitcnt Time, zDelay    'Wait
        xor outa, Pin           'Toggle Pin
        jmp #:zloop             'Loop endlessly
        
Pin     long |< 15              'Pin number
zDelay  long 15                 'Clock cycles to delay    - stops working if under 15 and a 4Mhz Z80
Time    res 1                   'System Counter Workspace
      

' ************************ I2C routines ***************************

PUB MCP23017_Out    ' uses Z80OUTA, Z80DIRA
' chips share the I2c bus with the eeprom. Pin 18 reset high, pin 19 and 20 = NC, A0-1-2 on first chip all low
' and on second chip A0 is high. The chips can be run from 3 or 5V (useful for level translation)
' gives a propeller another 32 input and output pins. Futurlec sell these chips for $1.20. DIP28 narrow package.
' if using sockets and can't source 28 pin sockets, use two 14 pin ones. 
' use johnny macs code, the io codes are different for the 23016 and 23017
  mcp23017.start(i2cSCL,i2cSDA,%000)                    ' clock, data, address of chip (2 chips, either 000 or 001)   
  mcp23017.cfg_dira(!(Z80DIRA & %11111111))             ' lsb. ! inverts bits. spin uses 1 for output and 23017 is the reverse so NOT the value
  mcp23017.wr_gpioa(Z80OUTA & %11111111)                ' output low byte 
  mcp23017.cfg_dirb(!(Z80DIRA >> 8))                    ' direction
  mcp23017.wr_gpiob(Z80OUTA >> 8)                       ' second lowest byte
  mcp23017.stop                                         ' stop this chip
  mcp23017.start(i2cSCL,i2cSDA,%001)                    ' clock, data, address of second chip   
  mcp23017.cfg_dira(!(Z80DIRA >> 16))                   ' direction
  mcp23017.wr_gpioa(Z80OUTA >> 16)                      ' output second high byte
  mcp23017.cfg_dirb(!(Z80DIRA >> 24))                   ' direction
  mcp23017.wr_gpiob(Z80OUTA >> 24)                      ' output high byte
  mcp23017.stop

PUB MCP23017_InA ' reads the A port on IC4
  mcp23017.start(i2cSCL,i2cSDA,%000)                    ' clock, data, address of chip (2 chips, either 000 or 001)   
  mcp23017.cfg_dira(%11111111)                          ' 1 = input
  result := mcp23017.rd_gpioa                           ' read the bits in port A
  mcp23017.stop   

VAR
  long Z80OUTA
  long Z80DIRA

 ' *********** Z80 I2C interface routines ********************
' three variables similar to spin/pasm Z80INA, Z80OUTA and Z80DIRA
' the DIRA variable defines which pins are inputs and which are outputs
' use the Spin protocol for DIRA; H = output, L = Input (the MCP23017 is the opposite, so ! inverts in the mcp23017_out code)

' Reset.Busrq.Mreq.Rd.Wr.3 Bank.16 address.8 data = 32 pins

'PUB Z80Start
'   Z80DIRA := %11000111_00000000_00000000_00000000      ' reset, busrq and bank pins outputs, rest all HiZ   
'   Z80ChangeBank(%000)                                  ' set to bank zero
'   Z80BusrqHigh                                         ' set busrq high
'   Z80ResetLow                                          ' reset low
'   delay.pause1ms(1)                                    ' short delay
'   Z80Run                                               ' sets reset high again
               

PUB Z80ResetLow
   Z80DIRA |= %10000000_00000000_00000000_00000000      ' set the direction of the reset pin to output
   Z80OUTA &= %01111111_11111111_11111111_11111111      ' set the output to Low
   MCP23017_Out
   
PUB Z80ResetHigh
   Z80DIRA |= %10000000_00000000_00000000_00000000      ' set the direction of the reset pin to output
   Z80OUTA |= %10000000_00000000_00000000_00000000      ' set the output to High
   MCP23017_Out

PUB Z80BusrqLow
   Z80DIRA |= %01000000_00000000_00000000_00000000      ' set the direction of the reset pin to output
   Z80OUTA &= %10111111_11111111_11111111_11111111      ' set the output to Low
   MCP23017_Out
   
PUB Z80BusrqHigh
   Z80DIRA |= %01000000_00000000_00000000_00000000      ' set the direction of the reset pin to output
   Z80OUTA |= %01000000_00000000_00000000_00000000      ' set the output to High
   MCP23017_Out   

PUB Z80Run                                              ' set pins so Z80 runs
   Z80OUTA := %01000000_00000000_00000000_00000000      ' reset low, busrq high, bank zero
   Z80DIRA := %11000111_00000000_00000000_00000000      ' reset, busrq and bank pins outputs, rest all HiZ
   MCP23017_Out
   delay.pause1ms(1)                                    ' for a proper reset
   Z80OUTA |= %10000000_00000000_00000000_00000000      ' reset pin high, Z80 can now run
   MCP23017_Out

PUB Z80Stop                                             ' set busrq low and wait for Z80 to respond
   Z80BusrqLow                                          ' note clock needs to be running for this to have any effect
   delay.pause1ms(1)                                    ' small delay to wait for Z80 to respond, and could be a lot shorter than this.
   Z80DIRA := %11111111_11111111_11111111_00000000      ' control all the pins except data
   Z80OUTA := %10111000_00000000_00000000_00000000      ' control pins high except busrq which is low. Bank is zero
   MCP23017_Out                                         ' output

PUB Z80ChangeBank(bank)                                 ' uses Z80Bank %000 to %111.    
   Z80OUTA &= %11111000_11111111_11111111_11111111      ' set bank to zero
   Z80OUTA |= (bank << 24)                              ' shift in the bank value and or 
   ' calling routine needs to output
  
PUB Z80WriteByte(address,data)                          ' address includes the bank so A0-A18, 512k.
   Z80DIRA := %11111111_11111111_11111111_11111111      ' all pins are outputs 
   Z80OUTA &= %11111000_00000000_00000000_00000000      ' clear the bank, address and data bits
   Z80OUTA |= (address << 8)                            ' merge z80outa with the address (including bank)
   Z80OUTA |= data                                      ' merge with data
   Z80OUTA &= %11010111_11111111_11111111_11111111      ' set the write and mreq pins low
   MCP23017_Out                                         ' output
   Z80OUTA |= %00101000_00000000_00000000_00000000      ' set the write and mreq pins high
   MCP23017_Out                                         ' output
   Z80DIRA := %11111111_11111111_11111111_00000000      ' data pins are now inputs (default condition) for next routine
   Z80OUTA &= %11111111_11111111_11111111_00000000      ' clear the data pins too 

PUB Z80ReadByte(address)                                ' address includes the bank so A0-A18 = 512k
   Z80DIRA := %11111111_11111111_11111111_00000000      ' all pins outputs except the data pins  
   Z80OUTA &= %11111000_00000000_00000000_00000000      ' clear the bank, address and data bits
   Z80OUTA |= (address << 8)                            ' merge z80outa with the address (including bank)
   Z80OUTA &= %11001111_11111111_11111111_11111111      ' set the read and mreq pins low
   MCP23017_Out                                         ' output
   result :=  MCP23017_InA                              ' get the byte
   Z80OUTA |= %00111000_00000000_00000000_00000000      ' set the read and mreq pins high
   MCP23017_Out                                         ' output

PUB Z80Dump(n) | i,j
   j := 0
   vt100.hex(0,4)               ' address
   vt100.out(32)
   repeat i from 0 to n
     vt100.hex(z80ReadByte(i),2)
     vt100.out(32)
     j +=1
     if j > 15
       j := 0                                    
       crlf
       vt100.hex(i+1,4) ' address
       vt100.out(32)
   crlf    
   

PUB Bootstrap | i ' send a tiny program capable of receiving a block of data
 ' uses 4 Z80 pins, 2 each way.
 ' P0 = Prop to Z80.
 ' P1 = Prop to Z80. 
 ' P4 = Z80 to Prop. 
 ' P5 = Z80 to Prop.
' write this in the N8VEM program and compile using RASM, view with hexedit, copy and paste from hexedit as hex text
{ 
        .Z80
        org 0h          ; need this to start at zero, add other org statements as needed

main:
        nop                     ; first byte ? gets corrupted to zero so put in a NOP
        ld      sp,0FFFFh       ; stack
        call    fetchbyte       ; get low byte of address to store
        ld      l,d             ; store in l
        call    fetchbyte       ; get high byte of address to store
        ld      h,d             ; store in h
        
mainloop:
        call    fetchbyte       ; get a byte in d
        ld      (hl),d  ; store in memory
        inc     hl              ; increment counter
        jr      mainloop        ; repeat

fetchbyte:                      ; returns byte in d
        ld      d,0             ; byte coming in is stored in d so reset to zero
        ld      b,8             ; load b with 8 to count the bits
get8bits:                       ; do 8 times
        call    fetchbit        ; waits till data is clocked in, returns e = 0 or 1
        ld      a,d             ; get previous d
        or      e               ; or with the returned bit
        ld      d,a             ; store back to d
        rrc     d               ; rotate right 
        djnz  get8bits  ; do this 8 times
        ret                     ; return

fetchbit:                       ; returns e = 0 or 1
        in      a,(0)           ; ports are not decoded so any port number will work
        bit     1,a             ; bit 0 is the data and bit 1 is the clock, wait till High
        jr      z,fetchbit      ; loop until the clock goes to 1
        and     00000001b       ; mask off other bits
        ld      e,a             ; store answer in e
clockzero:                      ; wait until the clock goes to zero again
        in      a,(0)           ; read the port         
        bit     1,a             ; test the clock bit, wait till Low
        jr      nz,clockzero; loop
        ret                     ; return
}
' see the dat section below for the compiled program

         printstringcr(string("Load Bootstrap"))
         Z80Stop
         repeat i from 0 to 63                          ' just a bit longer than the program in case change a few bytes
            Z80WriteByte(i,byte[@Z80BootProgram][i])
         DIRA |= %00000000_00000000_00000000_00000011   ' bit 0 is data and bit 1 is clock
         OUTA &= %11111111_11111111_11111111_11111100   ' both bits zero
         Z80Run                                         ' run this program on the Z80
         SendByte($40)                                  ' low byte of address to send data to
         SendByte($00)                                  ' high byte of address to send data to
         repeat 100
           SendByte($AB)                                  ' send out data
         Z80Stop
         Z80Dump(80)
DAT
   Z80BootProgram
        byte $00,$31,$FF,$FF,$CD,$13,$00,$6A,$CD,$13,$00,$62,$CD,$13,$00,$72 
        byte $23,$18,$F9,$16,$00,$06,$08,$CD,$22,$00,$7A,$B3,$57,$CB,$0A,$10 
        byte $F6,$C9,$DB,$00,$CB,$4F,$28,$FA,$E6,$01,$5F,$DB,$00,$CB,$4F,$20 
        byte $FA,$C9
                                   
PUB SendByte(n) ' send a byte out serially to the Z80
     repeat 8
       OUTA |= %00000000_00000000_00000000_00000010 | (n & %00000001) ' clock bit high, or with data bit
       repeat 3                                      ' with a 4Mhz Z80 clock, 2 works, 1 fails, use 3 to be safe
       OUTA &= %11111111_11111111_11111111_11111100 ' clock low, data low cleared for next bit
       repeat 3
       n >>= 1             
}
 