'' =================================================================================================
''
''   File....... jm_mcp23017.spin
''   Author..... Jon "JonnyMac" McPhalen (aka Jon Williams)
''               Copyright (c) 2011 Jon McPhalen
''               -- see below for terms of use
''   E-mail..... jon@jonmcphalen.com
''   Started.... 
''   Updated.... 
''
'' =================================================================================================


con

  MCP23017 = %0100_000_0                                        ' I2C slave ID for MCP23016

  ' command bytes for standard POR settings
  ' -- IOCON.BANK = 0
  
  IODIRA   = $00                                                ' command bytes 
  IODIRB   = $01 
  IPOLA    = $02 
  IPOLB    = $03 
  GPINTENA = $04 
  GPINTENB = $05 
  DEFVALA  = $06 
  DEFVALB  = $07 
  INTCONA  = $08 
  INTCONB  = $09 
  IOCON    = $0A 
  GPPUA    = $0C 
  GPPUB    = $0D 
  INTFA    = $0E 
  INTFB    = $0F 
  INTCAPA  = $10 
  INTCAPB  = $11 
  GPIOA    = $12 
  GPIOB    = $13 
  OLATA    = $14
  OLATB    = $15 


obj

  i2c : "jm_i2c_basic"

  
var

  long  slaveid                                                 ' 


pub start(sclpin, sdapin, address)

'' Start MCP23017 device; setup slave ID for this object

  i2c.setup(sclpin, sdapin)                                     ' connect to i2C
  slaveid := MCP23017 | ((address & %111) << 1)                 ' setup slave ID byte 


pub stop

'' Release this object fro I2C buss
'' -- buss pins remain active if other devices using I2C driver

  i2c.terminate


pub cfg_dira(dirbits)

'' Configure GPIOA IO pins
'' -- 0 bit = output, 1 bit = input

  i2c.start
  i2c.write(slaveid)
  i2c.write(IODIRA)
  i2c.write(dirbits)
  i2c.stop


pub cfg_pola(polbits)

'' Configure GP0 input polarity
'' -- 0 bit = normal, 1 bit = inverted

  i2c.start
  i2c.write(slaveid)
  i2c.write(IPOLA)
  i2c.write(polbits)
  i2c.stop


pub cfg_pua(pubits)

'' Configure GPIOA pull-ups
'' -- 1 bit = pull-up pin via 100K

  i2c.start
  i2c.write(slaveid)
  i2c.write(GPPUA)
  i2c.write(pubits)
  i2c.stop
  

pub wr_gpioa(outbits)

'' Write bits to GPIOA

  i2c.start
  i2c.write(slaveid)
  i2c.write(GPIOA)
  i2c.write(outbits)
  i2c.stop 


pub rd_gpioa | inbits

'' Read GPIOA bits

  i2c.start
  i2c.write(slaveid)
  i2c.write(GPIOA)
  i2c.start
  i2c.write(slaveid | %1)                                       ' read
  inbits := i2c.read(i2c#NAK)
  i2c.stop

  return inbits


pub cfg_dirb(dirbits)

'' Configure GPIOB IO pins
'' -- 0 bit = output, 1 bit = input

  i2c.start
  i2c.write(slaveid)
  i2c.write(IODIRB)
  i2c.write(dirbits)
  i2c.stop


pub cfg_polb(polbits)

'' Configure GPIOB input polarity
'' -- 0 bit = normal, 1 bit = inverted

  i2c.start
  i2c.write(slaveid)
  i2c.write(IPOLB)
  i2c.write(polbits)
  i2c.stop


pub cfg_pub(pubits)

'' Configure GPIOB pull-ups
'' -- 1 bit = pull-up pin via 100K

  i2c.start
  i2c.write(slaveid)
  i2c.write(GPPUB)
  i2c.write(pubits)
  i2c.stop
 

pub wr_gpiob(outbits)

'' Write value to GPIOB

  i2c.start
  i2c.write(slaveid)
  i2c.write(GPIOB)
  i2c.write(outbits)
  i2c.stop 


pub rd_gpiob | inbits

'' Read GPIOB bits

  i2c.start
  i2c.write(slaveid)
  i2c.write(GPIOB)
  i2c.start
  i2c.write(slaveid | %1)
  inbits := i2c.read(i2c#NAK)
  i2c.stop

  return inbits

 
dat

{{

  Terms of Use: MIT License

  Permission is hereby granted, free of charge, to any person obtaining a copy of this
  software and associated documentation files (the "Software"), to deal in the Software
  without restriction, including without limitation the rights to use, copy, modify,
  merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
  permit persons to whom the Software is furnished to do so, subject to the following
  conditions:

  The above copyright notice and this permission notice shall be included in all copies
  or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
  INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
  PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
  HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
  CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
  OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

}}